from datetime import datetime, timedelta
from typing import Optional

import cv2

from mower.solvers.navigation import NavigationSolver
from mower.solvers.navigation.utils import generate_name
from mower.utils import config
from mower.utils import typealias as tp
from mower.utils.datetime import get_server_weekday
from mower.utils.email import drop_template, send_message
from mower.utils.graph import SceneGraphSolver
from mower.utils.image import cropimg, diff_ratio, loadimg
from mower.utils.log import logger
from mower.utils.path import get_path
from mower.utils.recognize import Scene
from mower.utils.vector import va

drop_data = {}
for i in get_path("@install/ui/public/depot").iterdir():
    if "信物" in i.stem:
        continue
    supplement = "罗德岛物资补给"
    if i.stem.startswith(supplement) and i.stem != supplement:
        continue
    drop_data[i.stem] = loadimg.__wrapped__(i, True, "BLACK")
drop_digits = [generate_name(str(i), font_size=28, style="dark") for i in range(10)]


class OperationSolver(SceneGraphSolver):
    solver_name = "代理作战"

    def set_stop_time(self):
        if self.operation_start_time is None:
            operation_time = timedelta(minutes=3)
        else:
            operation_time = datetime.now() - self.operation_start_time
        if self.scheduler_stop_time is None:
            self.stop_time = None
        else:
            self.stop_time = self.scheduler_stop_time - operation_time

    def run(self):
        # 计算刷关时间和超时时间
        self.operation_start_time = None
        self.set_stop_time()

        self.sanity_drain = False
        self.timeout = False

        self.drop_list: list[dict[str, int]] = []
        return super().run()

    def number(self, scope: tp.Scope, height: Optional[int] = None):
        rect_limits = [{"w": 5, "h": 5, "char": ""}]
        return config.recog.num.number_int(
            "secret_front", scope, height, rect_limits=rect_limits
        )

    def drop_animation(self) -> bool:
        drop_scope = (100, 775), (1920, 945)
        img1 = cropimg(config.recog.gray, drop_scope)
        self.sleep()
        img2 = cropimg(config.recog.gray, drop_scope)
        return diff_ratio(img1, img2, ratio=0.03)

    def drop_recog(self):
        drop_result = {}
        for name, img in drop_data.items():
            score, scope = config.recog.matcher.match(img)
            if score < 0.7:
                continue
            logger.debug(f"{name=} {score=} {scope=}")
            target = config.recog.gray.copy()
            target = cropimg(target, scope)
            target = cropimg(target, ((0, 125), (180, 155)))

            values = []
            for _ in range(10):
                max_score, max_pos, max_digit = 0, None, None
                for n in range(10):
                    result = cv2.matchTemplate(
                        target, drop_digits[n], cv2.TM_CCOEFF_NORMED
                    )
                    min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
                    if max_val > max_score:
                        max_score, max_pos, max_digit = max_val, max_loc, n
                if max_score < 0.75:
                    break

                logger.debug(f"{max_score=} {max_pos=} {max_digit=}")
                values.append((max_pos[0], max_digit))
                h, w = drop_digits[max_digit].shape
                cv2.rectangle(target, max_pos, va(max_pos, (w, h)), (0,), -1)

            logger.debug(f"{name=} {values=}")

            values = sorted(values)
            value = 0
            for v in values:
                value = value * 10 + v[1]
            if value != 0:
                drop_result[name] = value

        logger.debug(f"{drop_result=}")
        self.drop_list.append(drop_result)

    def transition(self):
        if (scene := self.scene()) == Scene.OPERATOR_BEFORE:
            if config.recog.gray[65][1333] < 200:
                self.sleep()
                return
            if config.recog.gray[907][1600] < 127:
                self.tap((1776, 908))
                return
            repeat = self.number(((1520, 890), (1545, 930)), 28)
            if repeat > 1:
                self.tap((1500, 910))
                self.tap((1500, 801))
                return
            self.operation_start_time = datetime.now()
            self.tap_element("ope_start")
        elif scene == Scene.OPERATOR_SELECT:
            if self.stop_time and datetime.now() > self.stop_time:
                self.timeout = True
                return True
            self.tap((1655, 781))
        elif scene == Scene.OPERATOR_FINISH:
            self.set_stop_time()
            if self.drop_animation():
                return
            self.drop_recog()
            self.tap((310, 330))
        elif scene == Scene.OPERATOR_FAILED:
            self.tap((310, 330))
        elif scene == Scene.OPERATOR_ONGOING:
            if self.find("ope_agency_fail"):
                self.tap((121, 79))
            else:
                self.sleep(10)
        elif scene == Scene.OPERATOR_GIVEUP:
            self.tap_element("fight/give_up")
        elif scene == Scene.OPERATOR_RECOVER_POTION:
            use_medicine = False
            # 先看设置是否吃药
            if config.conf.maa_expiring_medicine:
                if config.conf.exipring_medicine_on_weekend:
                    use_medicine = get_server_weekday() >= 5
                else:
                    use_medicine = True
            # 再看是否有药可吃
            if use_medicine:
                img = cropimg(config.recog.img, ((1015, 515), (1170, 560)))
                img = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)
                img = cv2.inRange(img, (170, 0, 0), (174, 255, 255))
                count = cv2.countNonZero(img)
                logger.debug(f"{count=}")
                use_medicine = count > 3000
            if use_medicine:
                logger.info("使用即将过期的理智药")
                self.tap((1635, 865))
                return
            else:
                self.sanity_drain = True
                return True
        elif scene == Scene.OPERATOR_RECOVER_ORIGINITE:
            self.sanity_drain = True
            return True
        elif scene == Scene.OPERATOR_ELIMINATE:
            if self.find("ope_agency_lock"):
                logger.error("无法代理当期剿灭")
                return True
            if self.find("1800"):
                logger.info("本周剿灭已完成")
                return True
            if pos := self.find("ope_elimi_agency"):
                self.tap(pos)
                return
            self.operation_start_time = datetime.now()
            self.tap_element("ope_start")
        elif scene == Scene.OPERATOR_ELIMINATE_AGENCY:
            if self.find("can_not_use_ope"):
                logger.warning("当期剿灭代理获取合成玉数量为0")
                return True
            self.tap_element("ope_elimi_agency_confirm", interval=2)
        elif scene == Scene.UPGRADE:
            self.tap((960, 540))
        elif scene in self.waiting_scene:
            self.waiting_solver()
        else:
            return True


class OperationManager:
    solver_name = "刷理智"
    scheduler_stop_time: datetime | None = None

    def stage_today(self) -> list[str]:
        general_important = ["", "Annihilation"]
        general_unimportant = [
            "1-7",
            "LS-6",
            "CE-6",
            "AP-5",
            "SK-5",
            "CA-5",
            "PR-A-2",
            "PR-A-1",
            "PR-B-2",
            "PR-B-1",
            "PR-C-2",
            "PR-C-1",
            "PR-D-2",
            "PR-D-1",
        ]

        conf = config.conf
        plan_today = conf.weekly_plan[get_server_weekday()]
        result = []
        for stage in general_important:
            if stage in plan_today.general:
                result.append(stage)
        result += plan_today.custom
        for stage in general_unimportant:
            if stage in plan_today.general:
                result.append(stage)
        return result

    def run(self) -> bool:
        stage_today = self.stage_today()
        if task_finish := len(stage_today) == 0:
            send_message("今天没有要刷的关卡")
            return True

        nav_solver = NavigationSolver()
        ope_solver = OperationSolver()

        drop_list = []

        for name in stage_today:
            if self.scheduler_stop_time - datetime.now() < timedelta(minutes=3):
                break
            if nav_solver.run(name):
                ope_solver.scheduler_stop_time = self.scheduler_stop_time
                ope_solver.run()
                drop_list += ope_solver.drop_list
                if ope_solver.sanity_drain:
                    task_finish = True
                    break
                if ope_solver.timeout:
                    break

        if drop_list:
            drop_total = {}
            for drop in drop_list:
                for name, value in drop.items():
                    drop_total[name] = drop_total.get(name, 0) + value
            body = drop_template.render(drop_list=drop_list, drop_total=drop_total)
            send_message(body, "刷理智掉落")

        return task_finish
