from datetime import datetime, timedelta

import cv2

from mower.utils import config
from mower.utils.email import send_message
from mower.utils.graph import SceneGraphSolver
from mower.utils.image import cropimg
from mower.utils.log import logger
from mower.utils.recognize import Scene

from .sss_choose import SSSChooseSolver
from .sss_fight import SSSFightSolver
from .sss_navi import SSSNaviSolver
from .utils import generate_image, is_full


class SSSSolver(SceneGraphSolver):
    def run(self) -> bool:
        logger.info("Start: 保全作战")
        self.copilot_stages = [
            stage.model_dump(exclude_none=True) for stage in config.works["sss"].stages
        ]
        self.stages = {}
        if self.scheduler_stop_time:
            self.deadline = self.scheduler_stop_time - timedelta(minutes=1)
        else:
            self.deadline = None
        self.full = False  # 打满
        self.change_stages()
        super().run()
        return config.conf.sss.finish_while_full and self.full

    def change_stages(self):
        for stage in self.copilot_stages:
            stage_name = stage["stage_name"]
            del stage["stage_name"]
            self.stages[stage_name] = stage

    def detect_stage(self):
        for stage in self.stages:
            stage_name_size = 40 if len(stage) < 6 else 34
            res = generate_image(stage, stage_name_size)
            img = cropimg(config.recog.gray, ((180, 400), (420, 480)))
            result = cv2.matchTemplate(img, res, cv2.TM_CCOEFF_NORMED)
            min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
            logger.debug(f"stage:{stage},max_val:{max_val}")
            if max_val > 0.85:
                return stage
        return None

    def transition(self):
        if self.deadline and self.deadline < datetime.now():
            logger.info("任务打断保全")
            return True

        if (scene := self.scene()) in [Scene.SSS_DEPLOY, Scene.SSS_REDEPLOY]:
            stage = self.detect_stage()
            if SSSFightSolver().run(
                stage,
                self.stages[stage]["strategies"],
                self.stages[stage]["draw_as_possible"],
                self.stages[stage]["actions"],
                self.stages[stage]["retry_times"],
                deadline=self.deadline,
            ):
                return
            self.tap_element("sss/settlement_commission")

        elif (
            scene == Scene.SSS_MAIN and config.conf.sss.finish_while_full and is_full()
        ):
            send_message("保全模组条与仪已刷满", level="WARNING")
            self.full = True
            return True

        elif scene == Scene.SSS_DROP_AGENT_BEFORE_FIGHT:
            self.scene_graph_step(Scene.SSS_DEPLOY)

        elif scene == Scene.SSS_EXIT_CONFIRM:
            self.scene_graph_step(Scene.SSS_TERMINATED)

        elif scene == Scene.SSS_TERMINATED:
            logger.info("保全作战结束")
            return True

        elif scene == Scene.SSS_ACCOMPLISHED:
            logger.info("保全作战通关或终止")
            self.scene_graph_step(Scene.SSS_TERMINATED)

        elif scene == Scene.SSS_SQUAD:
            if config.conf.sss.choose_agent:
                SSSChooseSolver().run()
            self.tap_element("sss/squad")

        elif scene in self.waiting_scene:
            self.waiting_solver()
        else:
            SSSNaviSolver().run()
